import { getDocument, GlobalWorkerOptions, InvalidPDFException, MissingPDFException, UnexpectedResponseException, getPdfFilenameFromUrl } from 'pdfjs-dist/legacy/build/pdf'
import CryptoJS from 'crypto-js';
import printJS from 'print-js';
import { saveAs } from 'file-saver';
import { LinkTarget, PDFLinkService } from "./pdf_link_service"
import { PDFPrint, PDFPrintServiceFactory } from "./pdf_print_service"
import { PasswordPrompt } from "./password_prompt.js";
import { PDFFindBar } from "./pdf_find_bar.js";
import { PDFFindController } from "./pdf_find_controller.js";
import { PDFViewer } from './pdf_viewer'
import { PDFOutlineViewer } from "./pdf_outline_viewer.js";
import { PDFLayerViewer } from "./pdf_layer_viewer.js";
import { PDFThumbnailViewer } from './pdf_thumbnail_viewer'
import { PDFRenderingQueue } from "./pdf_rendering_queue";
import { PDFSidebar } from "./pdf_sidebar.js";
import { GenericExternalServices } from "./genericcom.js";
import { CursorTool, PDFCursorTools } from "./pdf_cursor_tools.js";
import { EventBus, isValidRotation, isValidScrollMode, shadow, isValidSpreadMode, toDateObject, findIndex, scrollIntoView, getElementsByTagName, DEFAULT_SCALE_VALUE, SidebarView, parseAnnotationFromXml } from './ui_utils'
import { RenderingStates, ANNOTATION_TYPE } from "../constants";
import { PDFPresentationMode } from "./pdf_presentation_mode.js";
import annotationStore from "./annotation_store"
import { InkSign } from "./ink_sign"

GlobalWorkerOptions.workerSrc = './lib/pdf.worker.min.js'
const CMAP_URL = './cmaps/'

class ComPDFKitViewer {
  #pwd = ''
  constructor (options) {
    this.config = options
    this.viewerContainer = null
    this.pdfLoadingTask = null
    this.printService = null
    this.pdfSidebar = null
    this.pdfOutlineViewer = null
    this.pdfLayerViewer = null
    this.pdfDocument = null
    this.downloadComplete = false
    this.isViewerEmbedded = window.parent !== window
    this.init = false
    this.isInitialViewSet = false
    this.pdfLinkService = null
    this.downloadManager = null
    this.externalServices = GenericExternalServices
    this.percent = 0
    this.pageChangedCallback = null
    this.scaleChangedCallback = null
    this.annotationsNumChangedCallback = null
    this.distanceChangedCallback = null
    this._license = ''
    this._token = null
    this._pdfId = null
    this.annotations = null
    this.annotationHistory = []
    this._annotationMode = 0
    this._boundEvents = Object.create(null)
    this._title = document.title
    this.activeTool = ''
    this.color = 'color'
    this.optionUrl = {
      baseUrl: 'https://test-compdf.kdan.cn',
      // baseUrl: 'https://wms.compdf.com',
      verifyUrl: '/api/license/verify',
      webviewBaseUrl: 'https://test-pdf-pro.kdan.cn:3032',
      // webviewBaseUrl: 'https://web-viewer-backend.compdf.com',
      uploadUrl: '/web-viewer/annotate/openPdfToAnnotate',
      editUrl: '/web-viewer/annotate/edit',
      saveUrl: '/web-viewer/annotate/save',
      uploadAnnotationsUrl: '/web-viewer/annotate/importAnnotateToPdf',
      downloadAnnotation: '/web-viewer/annotate/exportAnnotate',
      compareUrl: '/v1/pdf/compareto'
    }
    
    this.eventBus = new EventBus();
    this.toolMode = ''
  }

  get pagesCount() {
    return this.pdfDocument ? this.pdfDocument.numPages : 0;
  }

  get annotationsCount() {
    return this.pdfViewer.annotationStorage ? this.pdfViewer.annotationStorage.size : 0;
  }

  get page() {
    return this.pdfViewer.currentPageNumber;
  }

  set page(val) {
    this.pdfViewer.currentPageNumber = val;
  }

  get scale() {
    return this.pdfViewer.currentScale;
  }

  get supportsPrinting() {
    return PDFPrintServiceFactory.instance.supportsPrinting;
  }

  initConfig (options) {
    if (!options.license) return false
    this._license = options.license
    const data = {
      method: 'POST',
      headers: {
        'Content-Type': 'application/json'
      },
      body: JSON.stringify({
        license: this._license
      })
    }
    return fetch(this.optionUrl.baseUrl + this.optionUrl.verifyUrl, data)
      .then((res) => {
        return res.json()
      })
      .then((resp) => {
        if (resp.code === 200) {
          const verified = this._decrypt(resp.data)
          return verified
        } else {
          return false
        }
      })
      .catch((error) => ({ error: true, message: error.message }));
  }

  _decrypt (data) {
    const decrypt = CryptoJS.AES.decrypt(data, CryptoJS.enc.Utf8.parse('T2EKWNOflrd6ICU5'), {
      iv: CryptoJS.enc.Utf8.parse('GM3Faw9kX7CBmojt')
    })
    const decrypted = JSON.parse(decrypt.toString(CryptoJS.enc.Utf8))
    const hostname = window.location.hostname
    const reg = new RegExp(`^.*${decrypted.domain}$`);
    if (reg.test(hostname)) {
      this._token = decrypted.token
      return true
    }
    return false
  }

  initializeViewer ({
    container,
    viewer,
    thumbnailView,
    annotationView,
    outlineView,
    layerView,
    findbarView,
    toggleButton
  }) {
    this._initializeViewerComponents({
      container,
      viewer,
      thumbnailView,
      annotationView,
      outlineView,
      layerView,
      findbarView,
      toggleButton
    })
    this.bindEvents();
    this.bindWindowEvents();
  }

  setTool ({
    tool,
    color
  }) {
    this.activeTool = tool
    this.color = color
    this.eventBus.dispatch("toolChanged", { tool, color });
  }

  initAnnotations (annotations) {
    for (const type in annotations) {
      const annotation = annotations[type]
      if (Array.isArray(annotation)) {
        for (let i = 0; i < annotation.length; i++) {
          
          annotation[i].type = type
          this.handleAnnotations(annotation[i], true)
        }
      } else {
        annotation.type = type
        this.handleAnnotations(annotation, true)
      }
    }
    console.log(this.annotations)
    this.eventBus.dispatch('annotationChanged', { annotations: this.annotations })
  }

  handleAnnotations (annotation, init = false) {
    this.annotations = this.annotations || {}
    annotation.date = toDateObject(annotation.date)
    if (init) {
      this.initAddAnnotations(annotation)
    } else {
      if (Number(annotation.page) + 1 > this.pagesCount) return
      this.pushAnnotations(annotation)
    }
  }

  initAddAnnotations (annotation) {
    const page = annotation.page
    const annotations = this.annotations
    if (!annotations[page]) {
      annotations[page] = new Array(annotation.pageAnnotCount * 1)
      annotations[page][0] = {
        pageAnnotCount: 0
      }
    }
    const markup = ['highlight', 'underline', 'squiggly', 'strikeout']
    if (annotation.type !== 'link') {
      annotations[page][0].pageAnnotCount++
    }
    const pageAnnotCount = annotations[page][0].pageAnnotCount
    annotations[page][annotation.index] = annotation
    if (annotation.index == 0) {
      annotation.pageAnnotCount = pageAnnotCount
    }
    if (!annotationStore.annotationsAll) {
      annotationStore.annotationsAll = annotations
    }
  }

  pushAnnotations (annotation) {
    const page = annotation.page
    const annotations = this.annotations
    let pageAnnotCount = 0
    if (!annotations[page]) {
      pageAnnotCount = 1
      annotations[page] = []
      // debugger
      this.pdfViewer._pages[page].annotations = annotations[page]
    } else {
      pageAnnotCount = annotations[page][0].pageAnnotCount + 1
    }
    if (!annotationStore.annotationsAll) {
      annotationStore.annotationsAll = annotations
    }
    annotations[page].push(annotation)

    annotations[page][0].pageAnnotCount = pageAnnotCount
  }

  addAnnotations (annotation) {
    annotation.page -= 1
    this.handleAnnotationChange({
      type: 'add',
      annotation: [annotation]
    })
  }

  delAnnotations () {

  }

  addEvent (eventName, fn) {
    this.eventBus._on(eventName, fn);
  }

  async loadDocument (file, options = {}) {
    if (this.pdfLoadingTask) {
      await this.close()
    }
    if (options.annotations) {
      parseAnnotationFromXml(options.annotations)
    }
    this.info = options.info || null
    this.pageChangedCallback = options.pageChangedCallback || this.pageChangedCallback
    this.scaleChangedCallback = options.scaleChangedCallback || this.scaleChangedCallback
    this.annotationsNumChangedCallback = options.annotationsNumChangedCallback || this.annotationsNumChangedCallback
    this.distanceChangedCallback = options.distanceChangedCallback || this.distanceChangedCallback
    annotationStore.annotationsAll = null
    annotationStore.selectedName = null
    const parameters = {
      cMapUrl: CMAP_URL,
      cMapPacked: true,
      enableXfa: true
    };
    if (typeof file === "string") {
      // URL
      parameters.url = file
    } else if (file && "byteLength" in file) {
      // ArrayBuffer
      parameters.data = file;
    } else if (file.url && file.originalUrl) {
      parameters.url = file.url;
    }

    // 获取pdf信息，返回promise
    const loadingTask = getDocument(parameters)
    this.pdfLoadingTask = loadingTask
    this._docName = options.filename || getPdfFilenameFromUrl(parameters.url)
    const getPwd = (pwd) => this.#pwd = pwd
    loadingTask.onPassword = (updateCallback, reason) => {
      if (this.isViewerEmbedded) {
        // The load event can't be triggered until the password is entered, so
        // if the viewer is in an iframe and its visibility depends on the
        // onload callback then the viewer never shows (bug 1801341).
        this._unblockDocumentLoadEvent();
      }

      this.pdfLinkService.externalLinkEnabled = false;
      this.passwordPrompt.setUpdateCallback(updateCallback, reason, getPwd);
      this.passwordPrompt.open();
    };

    loadingTask.onProgress = ({ loaded, total }) => {
      const progress = options.progress || this.progress
      progress.call(this, loaded / total)
    }

    // 处理传入的PDF
    return await loadingTask.promise.then(
      async pdfDocument => {
        // 上传pdf
        // this.handleAnnotations(JSON.parse("{\"square\":{\"date\":\"2023-04-06_07:57:47\",\"rect\":\"159.707,339.463,337.691,452.407\",\"pageAnnotCount\":\"2\",\"color\":\"#00F900\",\"flags\":\"print\",\"name\":\"JMY5201X9645JIU20D1E78WBDHZH7CT7\",\"width\":\"2.000000\",\"index\":\"1\",\"page\":\"0\",\"opacity\":\"1.000000\",\"interior-opacity\":\"0.000000\"},\"line\":[{\"date\":\"2023-04-06_07:57:37\",\"pageAnnotCount\":\"2\",\"color\":\"#0000FF\",\"flags\":\"print\",\"start\":\"82.0325,553.659\",\"index\":\"0\",\"rect\":\"76.0325,547.659,275.837,671.854\",\"name\":\"KUF7565XS62841FVHN97U83X9C6J009W\",\"width\":\"5.000000\",\"end\":\"269.837,665.854\",\"interior-color\":\"#0000FF\",\"page\":\"0\",\"opacity\":\"1.000000\"},{\"date\":\"2023-04-06_07:57:41\",\"pageAnnotCount\":\"1\",\"color\":\"#0000FF\",\"flags\":\"print\",\"start\":\"205.61,511.057\",\"index\":\"0\",\"rect\":\"199.61,413.187,403.48,517.057\",\"name\":\"U3FZFYQT7WWD9I8AR4E80DLZ4T5N6Q3A\",\"width\":\"2.000000\",\"end\":\"397.48,419.187\",\"interior-color\":\"#0000FF\",\"page\":\"1\",\"opacity\":\"1.000000\"}]}"))
        await pdfDocument.getData().then(async (binaryData) => {
          const formData = new FormData();
          formData.append('file', new Blob([binaryData], { type: 'application/pdf' }), this._docName);
          formData.append('password', this.#pwd);
          const data = {
            method: 'POST',
            headers: {
              'Authorization': this._token
            },
            body: formData
          }
          await fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.uploadUrl, data)
            .then((res) => {
              return res.json()
            })
            .then((resp) => {
              if (resp.code === "200") {
                const data = resp.data
                this._pdfId = data.pdfId
                const annotations = data.annotateJson && JSON.parse(data.annotateJson)
                this.initAnnotations(annotations)
              } else if (resp.code === "319") {
                
              }
            })
            .catch((error) => {
              console.log(error)
              return { error: true, message: error.message }
            });
        })
        this.load(pdfDocument);
      },
      reason => {
        if (loadingTask !== this.pdfLoadingTask) {
          return undefined; // Ignore errors for previously opened PDF files.
        }

        let key = "loading_error";
        if (reason instanceof InvalidPDFException) {
          key = "invalid_file_error";
          alert('Invalid or corrupted PDF file.')
          return new Promise(function (resolve, reject) {
            reject('invalid_file_error')
          });
        } else if (reason instanceof MissingPDFException) {
          key = "missing_file_error";
          alert('Missing PDF file.')
        } else if (reason instanceof UnexpectedResponseException) {
          key = "unexpected_response_error";
          alert('Unexpected server response.')
        } else if (reason) {
          if (reason.message === 'No password given') {
            key = "no_password_given";
            alert('No password given.')
            return new Promise(function (resolve, reject) {
              reject('no_password_given')
            });
          } else if (reason.message === 'Incorrect Password') {
            return new Promise(function (resolve, reject) {
              reject('incorrect_password')
            });
          }
        }
        throw new Error(key)
      }
    )
  }

  setTitle (title = this._title) {

    if (this.isViewerEmbedded) {
      // Embedded PDF viewers should not be changing their parent page's title.
      return;
    }
    document.title = title;
  }

  get supportsFullscreen() {
    const fullscreenEnabled = document.webkitFullscreenEnabled || document.mozFullscreenEnabled || document.fullscreenEnabled
    return shadow(this, "supportsFullscreen", fullscreenEnabled);
  }

  removeAllAnnotations () {
    if (this.pdfViewer.annotationStorage) {
      this.pdfViewer.annotationStorage.cleanup()
    }
  }

  uploadFile(file) {
    return new Promise((resolve) => {
      const reader = new FileReader();
      reader.onload = () => {
        const contents = reader.result;
        resolve(contents);
      };
      reader.readAsText(file);
    })
  }

  async importAnnotations(data) {
    const annotationsXml = parseAnnotationFromXml(data)
    if (!annotationsXml.length) return
    this.handleAnnotationChange({
      type: 'add',
      annotation: annotationsXml
    })
    // for (let i = 0; i < len; i++) {
    //   const item = annotationsXml[i]
    //   const data = {
    //     annotationType: 3,
    //     fontSize: 10,
    //     value: item.obj_attr.content,
    //     rotation: 0,
    //     position: item.obj_attr.position,
    //     type: 'custom'
    //   }
    //   this.pdfViewer._pages[annotationsXml[i].obj_attr.page].pdfAnnotationLayer.deserialize(data)
    // }
  }

  saveAnnotations() {
    if (!this.pdfDocument || !this.info) return
    let append_objects = []
    for (let [key, value] of this.pdfViewer.annotationStorage._storage) {
      append_objects.push(value.annotationObj);
    }
    this.uploadAnnotations({
      append_objects,
      tid: this.info.tid
    })
  }

  initApiUrl(urls) {
    this.optionUrl = urls
  }

  getOptionUrl() {
    return this.optionUrl
  }

  uploadAnnotations(data) {
    const options = {
      method: 'POST',
      headers: {
        Accept: 'application/json',
        'Content-type': 'application/json',
      },
    }
    if (!data.append_objects.length) return
    options.body = JSON.stringify(data)
    fetch(this.optionUrl.baseUrl + this.optionUrl.saveAnnotations + '?f=' + this.info.token, options)
      .then((res) => {
        alert('Save Success')
        return res.json()
      })
      .then((resp) => resp)
      .catch((error) => ({ error: true, message: error.message }));
  }

  toggleSidebar () {
    this.pdfSidebar.toggle()
  }

  // 渲染PDF
  async load(pdfDocument) {
    this.pdfDocument = pdfDocument
    this.pdfLinkService.setDocument(pdfDocument);

    const pdfViewer = this.pdfViewer;
    pdfViewer.setDocument(pdfDocument, this.annotations);
    const { firstPagePromise, onePageRendered, pagesPromise } = pdfViewer;


    pdfDocument.getDownloadInfo().then(({ length }) => {
      this._contentLength = length; // Ensure that the correct length is used.
      this.downloadComplete = true;

      firstPagePromise.then(() => {
        this.eventBus.dispatch("documentloaded", { source: this });
      });
    })

    const pdfThumbnailViewer = this.pdfThumbnailViewer;
    pdfThumbnailViewer.setDocument(pdfDocument);

    
    firstPagePromise.then(pdfPage => {
      let hash = null;

      let rotation = null;
      let sidebarView = 0
      let scrollMode = -1
      let spreadMode = -1

      this.setInitialView(hash, {
        rotation,
        sidebarView,
        scrollMode,
        spreadMode,
      });
    })

    
    onePageRendered.then(data => {
      this.externalServices.reportTelemetry({
        type: "pageInfo",
        timestamp: data.timestamp,
      });

      pdfDocument.getOutline().then(outline => {
        if (pdfDocument !== this.pdfDocument) {
          return; // The document was closed while the outline resolved.
        }
        this.pdfOutlineViewer?.render({ outline, pdfDocument });
      });

      pdfViewer.optionalContentConfigPromise.then(optionalContentConfig => {
        if (pdfDocument !== this.pdfDocument) {
          return; // The document was closed while the layers resolved.
        }
        this.pdfLayerViewer?.render({ optionalContentConfig, pdfDocument });
      });
    });
    
    this.eventBus.dispatch("documentinit", { source: this });
    // Make all navigation keys work on document load,
    // unless the viewer is embedded in a web page.
    if (!this.isViewerEmbedded) {
      pdfViewer.focus();
    }
  }

  // 绑定自定义事件
  bindEvents() {
    const { eventBus, _boundEvents } = this;

    _boundEvents.beforePrint = this.beforePrint.bind(this);
    _boundEvents.afterPrint = this.afterPrint.bind(this);

    eventBus._on("resize", this.webViewerResize.bind(this));
    eventBus._on("pagerendered", this.webViewerPageRendered.bind(this));
    eventBus._on("beforeprint", _boundEvents.beforePrint);
    eventBus._on("afterprint", _boundEvents.afterPrint);
    eventBus._on("updateviewarea", this.webViewerUpdateViewarea.bind(this));
    eventBus._on("pagechanging", this.webViewerPageChanging.bind(this));
    eventBus._on("scalechanging", this.webViewerScaleChanging.bind(this));
    eventBus._on("rotationchanging", this.webViewerRotationChanging.bind(this));
    eventBus._on("sidebarviewchanged", this.webViewerSidebarViewChanged.bind(this));
    eventBus._on("pagemode", this.webViewerPageMode.bind(this));
    eventBus._on("namedaction", this.webViewerNamedAction.bind(this));
    eventBus._on("presentationmodechanged", this.webViewerPresentationModeChanged.bind(this));
    eventBus._on("presentationmode", this.webViewerPresentationMode.bind(this));
    eventBus._on("nextpage", this.webViewerNextPage.bind(this));
    eventBus._on("previouspage", this.webViewerPreviousPage.bind(this));
    eventBus._on("zoomin", this.webViewerZoomIn.bind(this));
    eventBus._on("zoomout", this.webViewerZoomOut.bind(this));
    eventBus._on("pagenumberchanged", this.webViewerPageNumberChanged.bind(this));
    eventBus._on("scalechanged", this.webViewerScaleChanged.bind(this));
    eventBus._on("rotatecw", this.webViewerRotateCw.bind(this));
    eventBus._on("rotateccw", this.webViewerRotateCcw.bind(this));
    eventBus._on("optionalcontentconfig", this.webViewerOptionalContentConfig.bind(this));
    eventBus._on("switchscrollmode", this.webViewerSwitchScrollMode.bind(this));
    eventBus._on("scrollmodechanged", this.webViewerScrollModeChanged.bind(this));
    eventBus._on("switchspreadmode", this.webViewerSwitchSpreadMode.bind(this));
    eventBus._on("spreadmodechanged", this.webViewerSpreadModeChanged.bind(this));
    eventBus._on("documentproperties", this.webViewerDocumentProperties.bind(this));
    eventBus._on("findfromurlhash", this.webViewerFindFromUrlHash.bind(this));
    eventBus._on("updatefindmatchescount", this.webViewerUpdateFindMatchesCount.bind(this));
    eventBus._on("updatefindcontrolstate", this.webViewerUpdateFindControlState.bind(this));

    eventBus._on("fileinputchange", this.webViewerFileInputChange.bind(this));
    eventBus._on("openfile", this.webViewerOpenFile.bind(this));
    eventBus._on("switchannotationeditormode", this.webViewerSwitchAnnotationEditorMode.bind(this));
    eventBus._on("annotationsCountChanged", this.webViewerAnnotationsCountChanged.bind(this))
    eventBus._on("distanceChanged", this.webViewerDistanceChanged.bind(this));
    eventBus._on("search", this.webViewerSearch.bind(this));

    eventBus._on("annotationChange", this.handleAnnotationChange.bind(this));
    eventBus._on("createSignature", this.handleCreateSignature.bind(this));
    eventBus._on("handleField", this.handleField.bind(this));
    eventBus._on("pageNumberChanged", this.pageNumberChanged.bind(this));
  }

  bindWindowEvents() {
    const { eventBus, _boundEvents } = this;

    _boundEvents.windowResize = () => {
      eventBus.dispatch("resize", { source: window });
    };
    _boundEvents.windowHashChange = () => {
      eventBus.dispatch("hashchange", {
        source: window,
        hash: document.location.hash.substring(1),
      });
    };
    _boundEvents.windowBeforePrint = () => {
      eventBus.dispatch("beforeprint", { source: window });
    };
    _boundEvents.windowAfterPrint = () => {
      eventBus.dispatch("afterprint", { source: window });
    };
    _boundEvents.windowUpdateFromSandbox = event => {
      eventBus.dispatch("updatefromsandbox", {
        source: window,
        detail: event.detail,
      });
    };

    window.addEventListener("resize", _boundEvents.windowResize);
    window.addEventListener("hashchange", _boundEvents.windowHashChange);
    window.addEventListener("beforeprint", _boundEvents.windowBeforePrint);
    window.addEventListener("afterprint", _boundEvents.windowAfterPrint);
    window.addEventListener(
      "updatefromsandbox",
      _boundEvents.windowUpdateFromSandbox
    );
  }

  async close () {
    if (!this.pdfLoadingTask) {
      return
    }
    const promises = [];

    promises.push(this.pdfLoadingTask.destroy());
    this.pdfLoadingTask = null
    if (this.pdfDocument) {
      this.pdfDocument = null;

      this.pdfThumbnailViewer.setDocument(null);
      this.pdfViewer.setDocument(null);
      this.pdfLinkService.setDocument(null);
    }
    this.pdfLinkService.externalLinkEnabled = true;
    this.isInitialViewSet = false;
    this.downloadComplete = false;
    this.url = "";
    this.baseUrl = "";
    this._downloadUrl = "";
    this.documentInfo = null;
    this.metadata = null;
    this._contentDispositionFilename = null;
    this._contentLength = null;
    this._saveInProgress = false;
    this._docStats = null;
    this._hasAnnotationEditors = false;

    this.pdfSidebar.reset();
    this.pdfOutlineViewer?.reset();
    this.pdfAttachmentViewer?.reset();
    this.pdfLayerViewer?.reset();

    this.#pwd = ''
    this._pdfId = null
    this.annotations = null

    if (this.findBar && this.findBar.opened) {
      this.findBar.close();
    }

    await Promise.all(promises);
  }

  handleInk (annotation) {
    let rawAnnotation = JSON.parse(JSON.stringify(annotation))
    let inklist=''
    let inkArray=[]
    if(Array.isArray(rawAnnotation.inklist))
    {
      inkArray=rawAnnotation.inklist
    }
    else
    {
      inkArray.push(rawAnnotation.inklist)
    }
    for(let i=0;i<inkArray.length;i++)
    {
      inklist=inklist+inkArray[i].replaceAll(';', ',')+ '//'
    }
    rawAnnotation.inklist =inklist// rawAnnotation.inklist.replaceAll(';', ',') + '//'
    return rawAnnotation
  }

  handleAnnotationChange (data) {
    const annotation = data.annotation
    this.annotationHistory.push(annotation)
    const annotateHandles = []

    const formType = ['text-field', 'check-box', 'radio-button', 'list-box', 'combo-box', 'push-button']

    if (data.type === 'add') {
      if (Array.isArray(annotation)) {
        const length = annotation.length
        // debugger
        for (let i = 0; i < length; i++) {
          if (Number(annotation[i].page) + 1 > this.pagesCount) continue
          if (this.pdfViewer) {
            this.pdfViewer.renderAnnotation(annotation[i])
          }
          annotation[i].operate = "add-annot"
          if (formType.includes(annotation[i].type)) {
            annotation[i].operate = 'add-form'
          }
          this.handleAnnotations(annotation[i])
          annotation[i].targetPage = annotation[i].page * 1 + 1
          let currentAnnotation = annotation[i]
          if (currentAnnotation.inklist) {
            currentAnnotation = this.handleInk(currentAnnotation)
          }
          annotateHandles.push(currentAnnotation)
        }
      } else {
        if (Number(annotation.page) + 1 > this.pagesCount) return
        this.handleAnnotations(annotation)

        annotation.targetPage = annotation.page + 1

        let currentAnnotation = annotation
        if (currentAnnotation.inklist) {
          currentAnnotation = this.handleInk(currentAnnotation)
        }
        annotateHandles.push(currentAnnotation)
      }
    } else {
      const annotations = this.annotations[annotation.page]
      const index = findIndex(annotation.name, annotations)
      if (data.type === 'delete') {
        const pageAnnotCount = --annotations[0].pageAnnotCount
        annotations.splice(index, 1)
        if (annotations.length) {
          annotations[0].pageAnnotCount = pageAnnotCount
        } else {
          delete this.annotations[annotation.page]
        }
      } else {
        const rawAnnotation = annotations[index]
        annotations[index] = {
          ...rawAnnotation,
          ...annotation
        }
      }
      annotation.index = index + 1
      annotation.targetPage = annotation.page + 1

      let currentAnnotation = annotation
      if (currentAnnotation.inklist) {
        currentAnnotation = this.handleInk(currentAnnotation)
      }
      annotateHandles.push(currentAnnotation)
    }

    this.eventBus.dispatch('annotationChanged', { annotations: this.annotations })

    const options = {
      method: 'POST',
      headers: {
        "Content-Type": "application/json",
        'Authorization': this._token
      },
      body: JSON.stringify({
        pdfId: this._pdfId,
        annotateHandles
      }),
    }
    fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.editUrl, options)
      .then((res) => {
        return res.json()
      })
      .then((data) => {
        if (data.code === "200") {
          return true
        } else {
          return false
        }
      })
      .catch((error) => (console.log(error)));
  }

  handleCreateSignature (data) {
    this.addAnnotations(data)
  }

  webViewerSearch (value) {
    this.findBar?.dispatchEvent("", false, value)
  }

  webViewerDistanceChanged (evt) {
    this.distanceChangedCallback(evt.distance)
  }

  progress (level) {
    if (this.downloadComplete) {
      // Don't accidentally show the loading bar again when the entire file has
      // already been fetched (only an issue when disableAutoFetch is enabled).
      return
    }
    const percent = Math.round(level * 100);
    // When we transition from full request to range requests, it's possible
    // that we discard some of the loaded data. This can cause the loading
    // bar to move backwards. So prevent this by only updating the bar if it
    // increases.
    if (percent <= this.percent) {
      return
    }
    this.percent = percent;
  }

  download (download = true) {
    if (this._token && this._pdfId) {
      const data = {
        method: 'POST',
        headers: {
          "Content-Type": "application/json",
          'Authorization': this._token
        },
        body: JSON.stringify({
          pdfId: this._pdfId
        })
      }
      return fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.saveUrl, data)
        .then((res) => {
          return res.json()
        })
        .then((data) => {
          if (data.code === "200") {
            if (download) {
              // this.downloadManager.downloadUrl(data.data.newFileUrl, this._docName)
              saveAs(data.data.newFileUrl, this._docName)
            }
            return data.data.newFileUrl
          } else {
            return false
          }
        })
        .catch((error) => {
          console.log(error)
          return false
        });
    } else {
      return false
    }
  }
  
  downloadXfdf () {
    return fetch(this.optionUrl.webviewBaseUrl + this.optionUrl.downloadAnnotation + "?pdfId=" + this._pdfId, {
      method: 'GET',
      headers: {
        "Content-Type": "application/json",
        'Authorization': this._token
      }
    })
      .then((res) => {
        return res.json()
      })
      .then((data) => {
        if (data.code === '200') {
          const name = this._docName.replace(/\.pdf$/, '.xfdf')
          // this.downloadManager.downloadUrl(data.data, name)
          saveAs(data.data, name)
          return true
        } else {
          return false
        }
      })
      .catch((error) => {
        console.log(error)
        return false
      });
  }

  _initializeViewerComponents ({
    container,
    viewer,
    thumbnailView,
    annotationView,
    outlineView,
    layerView,
    findbarView,
    toggleButton
  }) {
    this.viewerContainer = viewer
    const pdfRenderingQueue = new PDFRenderingQueue();
    pdfRenderingQueue.onIdle = this._cleanup.bind(this);
    this.pdfRenderingQueue = pdfRenderingQueue;

    const pdfLinkService = new PDFLinkService();
    this.pdfLinkService = pdfLinkService;

    const downloadManager = this.externalServices.createDownloadManager();
    this.downloadManager = downloadManager;

    const findController = new PDFFindController({
      linkService: pdfLinkService,
      eventBus: this.eventBus,
      updateMatchesCountOnProgress:
        typeof PDFJSDev === "undefined" || !PDFJSDev.test("GECKOVIEW"),
    });

    this.pdfViewer = new PDFViewer({
      container,
      viewer,
      annotationStore,
      tool: {
        tool: this.activeTool,
        color: this.color
      },
      annotationView,
      renderingQueue: pdfRenderingQueue,
      linkService: pdfLinkService,
      findController,
      renderer: 'canvas',
      textLayerMode: 1,
      enablePrintAutoRotate: true,
      eventBus: this.eventBus
    });

    pdfRenderingQueue.setViewer(this.pdfViewer);
    pdfLinkService.setViewer(this.pdfViewer);

    this.pdfThumbnailViewer = new PDFThumbnailViewer({
      container: thumbnailView,
      renderingQueue: pdfRenderingQueue,
      linkService: pdfLinkService
    });
    pdfRenderingQueue.setThumbnailViewer(this.pdfThumbnailViewer);

    this.findBar = new PDFFindBar({
      bar: document.getElementById("findbar"),
      toggleButton: document.getElementById("viewFind"),
      findField: document.getElementById("findInput"),
      highlightAllCheckbox: document.getElementById("findHighlightAll"),
      caseSensitiveCheckbox: document.getElementById("findMatchCase"),
      matchDiacriticsCheckbox: document.getElementById("findMatchDiacritics"),
      entireWordCheckbox: document.getElementById("findEntireWord"),
      findMsg: document.getElementById("findMsg"),
      findResultsCount: document.getElementById("findResultsCount"),
      findPreviousButton: document.getElementById("findPrevious"),
      findNextButton: document.getElementById("findNext"),
      findbarView
    }, this.eventBus);

    this.pdfCursorTools = new PDFCursorTools({
      container,
      eventBus: this.eventBus,
      cursorToolOnLoad: 0,
    });
    if (this.supportsFullscreen) {
      this.pdfPresentationMode = new PDFPresentationMode({
        container,
        pdfViewer: this.pdfViewer,
        eventBus: this.eventBus,
      });
    }

    this.passwordPrompt = new PasswordPrompt(
      {
        dialog: document.getElementById("passwordDialog"),
        label: document.getElementById("passwordText"),
        input: document.getElementById("password"),
        submitButton: document.getElementById("passwordSubmit"),
        cancelButton: document.getElementById("passwordCancel"),
      },
      this.isViewerEmbedded
    );

    this.pdfOutlineViewer = new PDFOutlineViewer({
      container: outlineView,
      eventBus: this.eventBus,
      linkService: pdfLinkService,
      downloadManager,
    });

    this.pdfLayerViewer = new PDFLayerViewer({
      container: layerView,
      eventBus: this.eventBus,
    });
  
    this.pdfSidebar = new PDFSidebar({
      elements: {
        toggleButton
      },
      pdfViewer: this.pdfViewer,
      pdfThumbnailViewer: this.pdfThumbnailViewer,
      eventBus: this.eventBus
    });
    this.pdfSidebar.onToggled = this.forceRendering.bind(this);
  }

  
  setInitialView(
    storedHash,
    { rotation, sidebarView, scrollMode, spreadMode } = {}
  ) {
    const setRotation = angle => {
      if (isValidRotation(angle)) {
        this.pdfViewer.pagesRotation = angle;
      }
    };
    const setViewerModes = (scroll, spread) => {
      if (isValidScrollMode(scroll)) {
        this.pdfViewer.scrollMode = scroll;
      }
      if (isValidSpreadMode(spread)) {
        this.pdfViewer.spreadMode = spread;
      }
    };
    this.isInitialViewSet = true;
    // this.pdfSidebar.setInitialView(sidebarView);

    setViewerModes(scrollMode, spreadMode);

    if (this.initialBookmark) {
      setRotation(this.initialRotation);
      delete this.initialRotation;

      this.pdfLinkService.setHash(this.initialBookmark);
      this.initialBookmark = null;
    } else if (storedHash) {
      setRotation(rotation);

      this.pdfLinkService.setHash(storedHash);
    }

    if (!this.pdfViewer.currentScaleValue) {
      // Scale was not initialized: invalid bookmark or scale was not specified.
      // Setting the default one.
      this.pdfViewer.currentScaleValue = DEFAULT_SCALE_VALUE;
    }
  }

  requestFullScreenMode() {
    this.pdfPresentationMode?.request();
  }

  /**
   * @private
   */
  _cleanup() {
    if (!this.pdfDocument) {
      return; // run cleanup when document is loaded
    }
    this.pdfViewer.cleanup();
    this.pdfThumbnailViewer.cleanup();
    // We don't want to remove fonts used by active page SVGs.
    this.pdfDocument.cleanup(
      /* keepLoadedFonts = */ this.pdfViewer.renderer === 'svg'
    );
  }

  cleanup () {
    this._cleanup()
  }

  zoomIn (steps) {
    if (this.pdfViewer.isInPresentationMode) {
      return;
    }
    this.pdfViewer.increaseScale(steps);
  }

  zoomOut (steps) {
    if (this.pdfViewer.isInPresentationMode) {
      return;
    }
    this.pdfViewer.decreaseScale(steps);
  }

  nextPage () {
    return this.pdfViewer.nextPage()
  }

  previousPage () {
    return this.pdfViewer.previousPage()
  }

  pageNumberChanged(value) {
    // Note that for `<input type="number">` HTML elements, an empty string will
    // be returned for non-number inputs; hence we simply do nothing in that case.
    if (value !== "") {
      this.pdfLinkService.goToPage(value);
    }

    // Ensure that the page number input displays the correct value, even if the
    // value entered by the user was invalid (e.g. a floating point number).
    if (
      value !== this.pdfViewer.currentPageNumber.toString() &&
      value !== this.pdfViewer.currentPageLabel
    ) {
      this.toolbar.setPageNumber(
        pdfViewer.currentPageNumber,
        pdfViewer.currentPageLabel
      );
    }
  }

  scrollTo ({
    left,
    top
  }) {
    scrollIntoView(this.viewerContainer, {
      left,
      top
    })
  }

  webViewerAnnotationsCountChanged(evt) {
    this.annotationsNumChangedCallback(evt.annotationsCount);
  }

  forceRendering() {
    this.pdfRenderingQueue.printing = !!this.printService;
    this.pdfRenderingQueue.isThumbnailViewEnabled =
      this.pdfSidebar.visibleView === SidebarView.THUMBS;
    this.pdfRenderingQueue.renderHighestPriority();
  }

  webViewerResize() {
    const { pdfDocument, pdfViewer, pdfRenderingQueue } = this;
  
    if (pdfRenderingQueue.printing && window.matchMedia("print").matches) {
      // Work-around issue 15324 by ignoring "resize" events during printing.
      return;
    }
    pdfViewer._updateContainerHeightCss();
  
    if (!pdfDocument) {
      return;
    }
    const currentScaleValue = pdfViewer.currentScaleValue;
    if (
      currentScaleValue === "auto" ||
      currentScaleValue === "page-fit" ||
      currentScaleValue === "page-width"
    ) {
      // Note: the scale is constant for 'page-actual'.
      pdfViewer.currentScaleValue = currentScaleValue;
    }
    pdfViewer.update();
  }

  webViewerPageRendered({ pageNumber, error }) {
    // If the page is still visible when it has finished rendering,
    // ensure that the page number input loading indicator is hidden.
    /** if (pageNumber === this.page) {
      this.toolbar.updateLoadingIndicatorState(false);
    } */
  
    // Use the rendered page to set the corresponding thumbnail image.
    if (this.pdfSidebar.visibleView === SidebarView.THUMBS) {
      const pageView = this.pdfViewer.getPageView(
        /* index = */ pageNumber - 1
      );
      const thumbnailView = this.pdfThumbnailViewer.getThumbnail(
        /* index = */ pageNumber - 1
      );
      if (pageView && thumbnailView) {
        thumbnailView.setImage(pageView);
      }
    }
  
    // if (error) {
    //   this.l10n.get("rendering_error").then(msg => {
    //     this._otherError(msg, error);
    //   });
    // }
  
    // It is a good time to report stream and font types.
    this._reportDocumentStatsTelemetry();
  }

  get annotationMode () {
    return this._annotationMode
  }

  set annotationMode (mode) {
    if (this.annotationMode === mode) {
      return
    }
    if (!this.pdfDocument) {
      return;
    }
    this._annotationMode = mode;
  }

  webViewerSwitchAnnotationEditorMode(evt) {
    this.pdfViewer.annotationEditorMode = evt.mode;
  }

  switchTool(mode) {
    this.pdfCursorTools.switchTool(mode);
  }

  _reportDocumentStatsTelemetry() {
    const { stats } = this.pdfDocument;
    if (stats !== this._docStats) {
      this._docStats = stats;

      // this.externalServices.reportTelemetry({
      //   type: "documentStats",
      //   stats,
      // });
    }
  }

  webViewerUpdateViewarea({ location }) {
    if (this.isInitialViewSet) {
      // Only update the storage when the document has been loaded *and* rendered.
      this.store
        ?.setMultiple({
          page: location.pageNumber,
          zoom: location.scale,
          scrollLeft: location.left,
          scrollTop: location.top,
          rotation: location.rotation,
        })
        .catch(() => {
          // Unable to write to storage.
        });
    }
    const href = this.pdfLinkService.getAnchorUrl(
      location.pdfOpenParams
    );
  
    // Show/hide the loading indicator in the page number input element.
    const currentPage = this.pdfViewer.getPageView(
      /* index = */ this.page - 1
    );
    const loading = currentPage?.renderingState !== RenderingStates.FINISHED;
    // this.toolbar.updateLoadingIndicatorState(loading);
  }
  
  webViewerScrollModeChanged(evt) {
    if (
      this.isInitialViewSet &&
      !this.pdfViewer.isInPresentationMode
    ) {
      // Only update the storage when the document has been loaded *and* rendered.
      this.store?.set("scrollMode", evt.mode).catch(() => {
        // Unable to write to storage.
      });
    }
  }
  webViewerScaleChanging(evt) {
    // this.toolbar.setPageScale(evt.presetValue, evt.scale);
    this.pdfViewer.update();
    this.scaleChangedCallback && this.scaleChangedCallback(evt.scale)
  }

  webViewerRotationChanging(evt) {
    this.pdfThumbnailViewer.pagesRotation = evt.pagesRotation;
  
    this.forceRendering();
    // Ensure that the active page doesn't change during rotation.
    this.pdfViewer.currentPageNumber = evt.pageNumber;
  }
  
  webViewerPageChanging({ pageNumber, pageLabel }) {
    // this.toolbar.setPageNumber(pageNumber, pageLabel);
    // this.secondaryToolbar.setPageNumber(pageNumber);
    if (this.pdfSidebar.visibleView === SidebarView.THUMBS) {
      this.pdfThumbnailViewer.scrollThumbnailIntoView(pageNumber);
    }
    this.pageChangedCallback(this.page)
  }
  
  webViewerSidebarViewChanged({ view }) {
    this.pdfRenderingQueue.isThumbnailViewEnabled =
      view === SidebarView.THUMBS;
  
    if (this.isInitialViewSet) {
      // Only update the storage when the document has been loaded *and* rendered.
      this.store?.set("sidebarView", view).catch(() => {
        // Unable to write to storage.
      });
    }
  }
  
  webViewerPageMode({ mode }) {
    // Handle the 'pagemode' hash parameter, see also `PDFLinkService_setHash`.
    let view;
    switch (mode) {
      case "thumbs":
        view = SidebarView.THUMBS;
        break;
      case "bookmarks":
      case "outline": // non-standard
        view = SidebarView.OUTLINE;
        break;
      case "attachments": // non-standard
        view = SidebarView.ATTACHMENTS;
        break;
      case "layers": // non-standard
        view = SidebarView.LAYERS;
        break;
      case "none":
        view = SidebarView.NONE;
        break;
      default:
        console.error('Invalid "pagemode" hash parameter: ' + mode);
        return;
    }
    this.pdfSidebar.switchView(view, /* forceOpen = */ true);
  }
  
  async webViewerNamedAction(evt) {
    // Processing a couple of named actions that might be useful, see also
    // `PDFLinkService.executeNamedAction`.
    switch (evt.action) {
      case "GoToPage":
        this.appConfig.toolbar.pageNumber.select();
        break;
  
      case "Find":
        if (!this.supportsIntegratedFind) {
          if (!evt.close || (evt.close && evt.close === 4 && this.findBar.opened)) {
            this.findBar.toggle();
          }
        }
        break;

      case "Print":
        const url = await this.download(false)
        this.triggerPrinting(url);
        return url
        break;
  
      case "SaveAs":
        this.downloadOrSave();
        break;
    }
  }

  triggerPrinting(url) {
    url = decodeURIComponent(url)
    printJS({
      url,
      printable: url,
      type: 'pdf',
      showModal: false
    });
  }

  webViewerPresentationModeChanged(evt) {
    this.pdfViewer.presentationModeState = evt.state;
  }
  
  webViewerPresentationMode() {
    this.requestPresentationMode();
  }
  
  webViewerNextPage() {
    this.pdfViewer.nextPage();
  }

  webViewerPreviousPage() {
    this.pdfViewer.previousPage();
  }
  webViewerZoomIn() {
    this.zoomIn();
  }
  webViewerZoomOut() {
    this.zoomOut();
  }

  webViewerPrint() {
    this.triggerPrinting();
  }

  webViewerPageNumberChanged(evt) {
    const pdfViewer = this.pdfViewer;
    // Note that for `<input type="number">` HTML elements, an empty string will
    // be returned for non-number inputs; hence we simply do nothing in that case.
    if (evt.value !== "") {
      this.pdfLinkService.goToPage(evt.value);
    }
  
    // Ensure that the page number input displays the correct value, even if the
    // value entered by the user was invalid (e.g. a floating point number).
    if (
      evt.value !== pdfViewer.currentPageNumber.toString() &&
      evt.value !== pdfViewer.currentPageLabel
    ) {
      this.toolbar.setPageNumber(
        pdfViewer.currentPageNumber,
        pdfViewer.currentPageLabel
      );
    }
  }
  webViewerScaleChanged(value) {
    this.pdfViewer.currentScaleValue = value;
  }
  webViewerRotateCw() {
    this.rotatePages(90);
  }
  webViewerRotateCcw() {
    this.rotatePages(-90);
  }
  webViewerOptionalContentConfig(evt) {
    this.pdfViewer.optionalContentConfigPromise = evt.promise;
  }
  webViewerSwitchScrollMode(mode) {
    this.pdfViewer.scrollMode = mode;
  }
  webViewerSwitchSpreadMode(mode) {
    this.pdfViewer.spreadMode = mode;
  }

  
  beforePrint() {

    if (this.printService) {
      // There is no way to suppress beforePrint/afterPrint events,
      // but PDFPrintService may generate double events -- this will ignore
      // the second event that will be coming from native window.print().
      return;
    }

    // The beforePrint is a sync method and we need to know layout before
    // returning from this method. Ensure that we can get sizes of the pages.
    if (!this.pdfViewer.pageViewsReady) {
      // eslint-disable-next-line no-alert
      window.alert('Warning: The PDF is not fully loaded for printing.');
      return;
    }

    const pagesOverview = this.pdfViewer.getPagesOverview();
    const printContainer = document.getElementById("printContainer");
    const printResolution = 150;
    const optionalContentConfigPromise =
      this.pdfViewer.optionalContentConfigPromise;
    const printService = PDFPrintServiceFactory.instance.createPrintService(
      this.pdfDocument,
      pagesOverview,
      printContainer,
      printResolution,
      optionalContentConfigPromise
    );
    this.printService = printService;
    this.forceRendering();
    this.setTitle(this._docName);

    printService.layout();

  }

  afterPrint() {
    if (this.printService) {
      this.printService.destroy();
      this.printService = null;

      this.pdfDocument?.annotationStorage.resetModified();
    }
    this.forceRendering();
    this.setTitle(this._title);
  }

  rotatePages(delta) {
    this.pdfViewer.pagesRotation += delta;
    // Note that the thumbnail viewer is updated, and rendering is triggered,
    // in the 'rotationchanging' event handler.
  }
  
  webViewerSpreadModeChanged(evt) {
    if (
      this.isInitialViewSet &&
      !this.pdfViewer.isInPresentationMode
    ) {
      // Only update the storage when the document has been loaded *and* rendered.
      this.store?.set("spreadMode", evt.mode).catch(() => {
        // Unable to write to storage.
      });
    }
  }
  webViewerDocumentProperties() {
    this.pdfDocumentProperties.open();
  }

  webViewerFindFromUrlHash(evt) {
    this.eventBus.dispatch("find", {
      source: evt.source,
      type: "",
      query: evt.query,
      phraseSearch: evt.phraseSearch,
      caseSensitive: false,
      entireWord: false,
      highlightAll: true,
      findPrevious: false,
      matchDiacritics: true,
    });
  }
  
  webViewerUpdateFindMatchesCount({ matchesCount }) {
    if (this.supportsIntegratedFind) {
      this.externalServices.updateFindMatchesCount(matchesCount);
    } else {
      this.findBar.updateResultsCount(matchesCount);
    }
  }
  
  webViewerUpdateFindControlState({
    state,
    previous,
    matchesCount,
    rawQuery,
  }) {
    if (this.supportsIntegratedFind) {
      this.externalServices.updateFindControlState({
        result: state,
        findPrevious: previous,
        matchesCount,
        rawQuery,
      });
    } else {
      this.findBar.updateUIState(state, previous, matchesCount);
    }
  }
  
  webViewerFileInputChange (evt) {
    if (this.pdfViewer?.isInPresentationMode) {
      return; // Opening a new PDF file isn't supported in Presentation Mode.
    }
    const file = evt.fileInput.files[0];
  
    let url = URL.createObjectURL(file);
    if (file.name) {
      url = { url, originalUrl: file.name };
    }
    this.open(url);
  };
  
  // eslint-disable-next-line no-var
  webViewerOpenFile (evt) {
    const fileInput = this.appConfig.openFileInput;
    fileInput.click();
  }

  setToolMode (mode) {
    this.toolMode = mode;
    this.eventBus.dispatch("toolModeChanged", mode);

    for (let page in this.annotations) {
      for (let i = 0; i < this.annotations[page].length; i++) {
        const annot = this.annotations[page][i]
        const element = document.getElementById(annot['name'])

        if (element) {
          if (this.toolMode !== 'form') {
            if (annot['isHidden'] === '1') {
              element.style.display = 'none';
              element.style.pointerEvents = 'none';
            } else if (annot['isHidden'] === '0') {
              element.style.display = 'block';
              element.style.pointerEvents = 'auto';
            }
          } else {
            element.style.display = 'block';
            element.style.pointerEvents = 'auto';
          }
        }
      }
    }
  }

  setPropertyPanel (props) {
    this.eventBus.dispatch("propertyPanelChanged", props);
  }

  setInitProperty () {
    if (this.activeTool === 'textfield') {
      this.eventBus.dispatch("initTextfieldName");
    } else if (this.activeTool === 'checkbox') {
      this.eventBus.dispatch("initCheckboxName");
    } else if (this.activeTool === 'radiobutton') {
      this.eventBus.dispatch("initRadioButtonName");
    } else if (this.activeTool === 'pushbutton') {
      this.eventBus.dispatch("initPushButtonName");
    }
  }

  handleField (prop) {
    const id = prop.id
    const key = prop.key
    const value = prop.value

    if (!id) return

    let annotationsAll = Object.values(annotationStore.annotationsAll).reduce(function(acc, curr) {
      return acc.concat(curr);
    }, []);
    let targetObj = annotationsAll.find(obj => obj.name === id)
    console.log(targetObj)

    if (key && value) {
      let props = {
        name: id
      }
      props[key] = value
      this.eventBus.dispatch("setProperty", props)
      console.log(annotationStore.annotationsAll)
    }
    return targetObj
  }

  handleSign (flag,param) {
    if (flag === 'create' && !this.InkSign) {
      this.InkSign = new InkSign({
        pdfViewer: this.pdfViewer,
        eventBus: this.eventBus,
      })
    }
    if (!this.InkSign) return;
    if (flag === 'clear' || flag === 'create') {
      this.InkSign.clear()
    }
    if (flag === 'save') {
      this.InkSign.toImage(param)
    }
    if (flag === 'reset') {
      this.InkSign.reset()
      this.InkSign = null
    }
    if(flag==="update")
    {
      this.InkSign.update(param)
    }
  }

  handleStamp (data) {
    if (!data) return
    console.log(data)
    let imgUi = data.stampContainer
    let pageNum = data.page
    let pageView = this.pdfViewer.getPageView(pageNum - 1)
    let rect = ''
    let annotData = null

    if (pageView && imgUi) {
      let pageWidth = pageView.viewport.viewBox[2]
      let pageHeight = pageView.viewport.viewBox[3]
      let imgWidth = imgUi.width
      let imgHeight = imgUi.height

      let left = (pageWidth - imgWidth) / 2
      let top = (pageHeight - imgHeight) / 2
      let right = (pageWidth + imgWidth) / 2
      let bottom = (pageHeight + imgHeight) / 2

      rect = left + ',' + top + ',' + right + ',' + bottom
    }
    if (data.stampType === 'standard') {
      annotData = {
        operate: 'add-annot',
        type: 'stamp',
        page: pageNum,
        rect,
        stampType: 'standard',
        standardStampType: imgUi.alt
      }
      this.addAnnotations(annotData)
    }
    if (data.stampType === 'dynamic') {
      annotData = {
        operate: 'add-annot',
        type: 'stamp',
        page: pageNum,
        rect,
        stampType: 'text',
        textStampFirststring: data.text,
        textStampSecondstring: data.time,
        textStampColor: data.color,
        textStampShape: data.shape
      }
      this.addAnnotations(annotData)
    }
  }
}

function webViewerTouchStart(evt) {
  if (
    PDFViewerApplication.pdfViewer.isInPresentationMode ||
    evt.touches.length < 2
  ) {
    return;
  }
  evt.preventDefault();

  if (evt.touches.length !== 2) {
    PDFViewerApplication._touchInfo = null;
    return;
  }

  let [touch0, touch1] = evt.touches;
  if (touch0.identifier > touch1.identifier) {
    [touch0, touch1] = [touch1, touch0];
  }
  PDFViewerApplication._touchInfo = {
    touch0X: touch0.pageX,
    touch0Y: touch0.pageY,
    touch1X: touch1.pageX,
    touch1Y: touch1.pageY,
  };
}

export default ComPDFKitViewer
